#ifndef _DENOISER_HPP_
#define _DENOISER_HPP_

#include <parameters.h>

#define RESIDUAL_THRESHOLD 0.1f

/*!
 * \class Denoiser
 *
 * \brief Restores a frame
 *
 * This class restores noisy pixels of a frame according to some noise-specific restoration kernel.
 */
class Denoiser {
public:
    Denoiser(unsigned int height_, unsigned int width_, bool fixed_cycles_, unsigned int max_cycles_, bool trace_time_) :
        height(height_), width(width_), fixed_cycles(fixed_cycles_), max_cycles(max_cycles_), out(NULL),
        trace_time(trace_time_), svc_time_us(0) {
    }

    /*!
     * \brief Returns the restored frame
     *
     * It returns the restored frame built by restore() method.
     *
     * \return the restore frame
     */
    unsigned char *getRestored() {
        return out;
    }

    unsigned long getSvcTime() {
        return svc_time_us;
    }

    /*!
     * \brief Restores a frame
     *
     * It iterates the restoration kernel over a frame until
     * convergence is reached.
     *
     * \param in is the input frame
     * \param noisymap is the noisy-map
     * \param noisy is the array of noisy pixels
     * \param n_noisy is the length of the array of noisy pixels
     *
     * \return the number of restoration cycles
     */
    unsigned int restore(unsigned char *in, int *noisymap, unsigned int *noisy, unsigned int n_noisy) {
        unsigned long timestamp_us;
        if(trace_time)
            timestamp_us = get_usec_from(0);
        out = (unsigned char *) malloc(height * width * sizeof(unsigned char));
        unsigned char *diff = (unsigned char *) malloc(n_noisy * sizeof(unsigned char));
        memset(diff, 0, n_noisy * sizeof(unsigned char));
        memcpy(out, in, height * width * sizeof(unsigned char));
        float *residuals = (float *) malloc(n_noisy * sizeof(float));
        bool fixed = false;
        unsigned int restore_cycles = 0;
        while (true) {
            //restore
            for (unsigned int i = 0; i < n_noisy; ++i) {
                unsigned int x = noisy[i];
                out[x] = restorePixel(in, x, noisymap);
                unsigned char newdiff = (unsigned char) (_ABS((int) (out[x]) - noisymap[x]));
                residuals[i] = (float) (_ABS((int) newdiff - (int) (diff[i])));
                diff[i] = newdiff;
            }
            //reduce residuals
            float residual = 0.0f;
            for (unsigned int i = 0; i < n_noisy; ++i)
                residual += residuals[i];
            residual /= n_noisy;
            ++restore_cycles;
            //check convergence
            if (fixed_cycles)
                fixed = restore_cycles == max_cycles;
            else
                fixed = residual < RESIDUAL_THRESHOLD || restore_cycles >= max_cycles;
            if (fixed)
                break;
            memcpy(in, out, height * width * sizeof(unsigned char));
        }
        //clean-up
        free(diff);
        free(residuals);
        free(noisymap);
        if(trace_time)
            svc_time_us += get_usec_from(timestamp_us);

        return restore_cycles;
    }

    /*!
     * \brief Restores a pixel
     *
     * It applies the restoration kernel over a pixel.
     *
     * \param in is the input frame
     * \param x is the index of the pixel to be restored
     * \param noisymap is the noisy-map
     *
     * \return the value of the restored pixel
     */
    virtual unsigned char restorePixel(unsigned char *in, unsigned int x, int *noisymap) = 0;

    virtual ~Denoiser() {}

protected:
    unsigned int height, width;
    bool fixed_cycles;
    unsigned int max_cycles;
    unsigned char *out;
    bool trace_time;
    unsigned long svc_time_us;

};
#endif
