#ifndef _DETECTOR_HPP_
#define _DETECTOR_HPP_

/*!
 * \class Detector
 *
 * \brief Detects noisy pixels
 *
 * This class detects noisy pixels according to some noise-specific detection kernel.
 */
class Detector {
public:

    Detector(unsigned int height_, unsigned int width_, bool trace_time_) :
        height(height_), width(width_), n_noisy(0), noisyMap(NULL), noisyPixels(NULL),
        trace_time(trace_time_), svc_time_us(0) {
    }

    /*!
     * \brief Returns the noisy-map
     *
     * It returns the noisy-map built by detect() method.
     *
     * \return the noisy-map. The value of the pixel
     * (x,y) of the noisy-map is -1 if the corresponding
     * pixel is not noisy, otherwise it has the same value
     * as the original frame
     */
    int * getNoisyMap() {
        return noisyMap;
    }

    /*!
     * \brief Returns the array of noisy pixels
     *
     * It returns the array of noisy pixels built by detect() method.
     *
     * \return the array of (indexes of) noisy pixels
     */
    unsigned int * getNoisyPixels() {
        return noisyPixels;
    }

    /*!
     * \brief Returns the number of noisy pixels
     *
     * It returns the number of noisy pixels.
     *
     * \return the number of noisy pixels
     */
    unsigned int getNbNoisy() {
        return n_noisy;
    }

    unsigned long getSvcTime() {
        return svc_time_us;
    }

    /*!
     * \brief Detects noise over a frame
     *
     * It applies the AMF over a frame.
     *
     * \param noisyImage is the input frame
     */
    void detect(unsigned char * noisyImage) {
        unsigned long timestamp_us;
        if(trace_time)
            timestamp_us = get_usec_from(0);
        init(noisyImage);
        
        //noisymap
        noisyMap = (int *) malloc(height * width * sizeof(int));
        //[[spar::ToStream, spar::Input(noisyImage)]]
        for (unsigned int ri = 0; ri < height; ++ri){
            //[[spar::Stage,spar::Input(noisyImage,ri),spar::Replicate(2)]]{
            for (unsigned int ci = 0, x = ri * width; ci < width; ++ci, ++x)
                noisyMap[x] = (isPixelNoisy(noisyImage, noisyImage[x], ri, ci)) ? noisyImage[x] : -1;
            //}
        }
        //array of noisy pixels
        n_noisy = 0;
        for (unsigned int ri = 0; ri < height; ++ri)
            for (unsigned int ci = 0, x = ri * width; ci < width; ++ci, ++x)
                if (noisyMap[x] >= 0)
                    ++n_noisy;
        noisyPixels = (unsigned int *) malloc(n_noisy * sizeof(unsigned int));
        for (unsigned int i = 0, ri = 0; ri < height; ++ri)
            for (unsigned int ci = 0, x = ri * width; ci < width; ++ci, ++x)
                if (noisyMap[x] >= 0)
                    noisyPixels[i++] = x;
        if(trace_time)
            svc_time_us += get_usec_from(timestamp_us);
    }

    /*!
     * \brief Initializes the detection engine
     *
     * It initializes the detection engine for a frame.
     *
     * \param noisyImage is the input frame
     */
    virtual void init(unsigned char *noisyImage) {
    }

    /*!
     * \brief Applies the detection kernel on a pixel
     *
     * It checks if a single pixel is noisy.
     *
     * \return
     * \retval TRUE if the pixel is noisy
     * \retval FALSE otherwise
     */
    virtual bool isPixelNoisy(unsigned char *im, unsigned char center, unsigned int r, unsigned int c) = 0;

    virtual ~Detector() {
    }
    ;

protected:
    unsigned int width;
    unsigned int height;
    int * noisyMap;
    unsigned int * noisyPixels;
    unsigned int n_noisy;
    bool trace_time;
    unsigned long svc_time_us;
};
#endif
