#ifndef _DETECTORSPD_HPP_
#define _DETECTORSPD_HPP_

#include <defs.h>
#include <Detector.hpp>

/*!
 * \class DetectorSPD
 *
 * \brief Detects salt-and-pepper noisy pixels
 *
 * This class detects salt-and-pepper noisy pixels
 * by applying a classical AMF.
 */
class DetectorSPD: public Detector {
public:

  /*!
   * \brief Creates instance of DetectorSPD
   *
   * It creates an instance of DetectorSPD.
   *
   * \param height is the height of the frame
   * \param width is the width of the frame
   * \param w_max_ is the max size of the window for the adaptive filter
   * \param trace_time_ tells if time should be measured
   */
  DetectorSPD(unsigned int height, unsigned int width, unsigned int w_max_, bool trace_time_) :
      Detector(height, width, trace_time), w_max(w_max_) {
  }

  /*!
   * \brief Check if a pixel is noisy
   *
   * It applies the AMF over a single pixel.
   *
   * \param im is the input frame
   * \param center is the value of the pixel
   * \param r is the row-coordinate of the pixel
   * \param c is the column-coordinate of the pixel
   *
   * \return true if the pixel is noisy
   */
  bool isPixelNoisy(unsigned char *im, unsigned char center, unsigned int r, unsigned int c) {
    return is_noisy(center, r, c, w_max, im, width, height);
  }

private:
  unsigned int w_max;

  int ctrl_px(unsigned char *im, unsigned char *a, unsigned int w, unsigned int yy, unsigned int xx, unsigned int width,
      unsigned int height) {
    unsigned int cw(w / 2);
    int n = 0;
    for (unsigned int r = 0; r < w; r++) //rows
      for (unsigned int c = 0; c < w; c++) //columns
        if (((xx + c) >= cw) && ((xx + c) < (width + cw)) && ((yy + r) >= cw) && ((yy + r) < (height + cw)))
          a[n++] = im[(yy + r - cw) * width + xx + c - cw];
    return n;
  }

  //AMF filter on grayscale pixel
  inline bool is_noisy(unsigned char center, unsigned int r, unsigned int c, unsigned int w_max, unsigned char *im, unsigned int width,
      unsigned int height) {
    if (center == SALT || center == PEPPER) {
      unsigned char c_array[MAX_WINDOW_SIZE * MAX_WINDOW_SIZE];
      for (unsigned int ws = 3; ws < w_max; ws += 2) {
        int ac = ctrl_px(im, c_array, ws, r, c, width, height);
        sort(c_array, c_array + ac);
        unsigned char min_cp = c_array[0];
        unsigned char med_cp = c_array[ac / 2];
        unsigned char max_cp = c_array[ac - 1];
        if ((!(min_cp == med_cp && med_cp == max_cp)) || (ws > w_max)) //not homogeneous zone
          if (center == min_cp || center == max_cp || ws > w_max)
            return true; //noisy
      }
    }
    return false; //not noisy
  }
};
#endif
